==> benno-web-2.10.7.tgz: srv/benno/bennoweb/htdocs/model/Backend.php <==
<?php
/*
 * Benno MailArchiv
 *
 * Copyright  2008-2012 LWsystems GmbH & Co. KG
 *
 * http://www.lw-systems.de/
 * http://www.benno-mailarchiv.de/
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3
 * as published by the Free Software Foundation.
 *
 * Binary versions of this file provided by LWsystems to you as
 * well as other copyrighted, protected or trademarked materials like
 * logos, graphics, fonts, specific documentations and configurations,
 * cryptographic keys etc. are subject to a license agreement between
 * you and LWsystems.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You can find a copy of the GNU Affero General Public License at
 * this URI:
 * http://www.gnu.org/licenses/agpl-3.0.html
 * If not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 */

require_once 'model/EMail.php';
require_once 'model/Attachment.php';
require_once 'model/QueryResult.php';
require_once 'model/License.php';

class Backend implements Serializable
{

    var $base_url;
    var $filter;        // filter assign to each query
    var $rest_user;
    var $rest_pass;
    var $rest_ssl_verify;
    var $_curl;

    function __construct($base_url,$rest_user,$rest_pass,$filter)
    {
        global $App;
        $this->rest_user = $rest_user;
        $this->rest_pass = $rest_pass;
        $this->rest_ssl_verify = $App->getConfig('REST_SSL_VERIFY','NO');

        $this->base_url = $base_url;
        $this->filter = $filter;
    }



    public function serialize()
    {
        $odata['base_url']        = $this->base_url;
        $odata['filter']          = $this->filter;
        $odata['rest_pass']       = $this->rest_pass;
        $odata['rest_ssl_verify'] = $this->rest_ssl_verify;
        $odata['rest_user']       = $this->rest_user;

        return serialize($odata);
    }


    public function unserialize($sdata)
    {
        $odata = unserialize($sdata);
        $this->base_url         = $odata['base_url'];
        $this->filter           = $odata['filter'];
        $this->rest_pass        = $odata['rest_pass'];
        $this->rest_ssl_verify  = $odata['rest_ssl_verify'];
        $this->rest_user        = $odata['rest_user'];

        $this->_curl = curl_init();
    }


    /**
     * Fetch one email from backend
     *
     * @param string $archive
     * @param string $sessionMailId
     * @return object $EMail
     */
    function fetch ($cidstring,$sessionMailId)
    {
        global $App;
        $bennoId = QueryResult::getBennoId($sessionMailId);
        $App->log("Fetch mail from archive: $bennoId",3,'Backend');

        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);
        $query_string = 'archive='.$archive.
                        '&id='.$bennoId;
        if ($scid)  $query_string .= '&scid='.$scid;

        $maildata = $this->backend_request('/mail',$query_string);

        $body   = $maildata['body'];
        $header = $maildata['header'];

        $EMail = new EMail($sessionMailId,$archive,$bennoId,$header['From'],$header['To'],$header['Cc'],$header['Bcc'],$header['Subject'],$header['Date'],$header['Size'],$body);
        $EMail->verified = $header['X-Benno-HashCheckPassed'];
        $EMail->dateTS   = $header['X-Benno-DateTS'];

        foreach ($maildata['attachments'] as $atm) {
            //$Attachment = new Attachment($atm['name'],@$atm['content-type'], $atm['size']);
        	$Attachment = new Attachment($atm['name'],@$atm['content-type'], $atm['size'], true, $atm['attachmentId']);
            $EMail->addAttachment($Attachment);
        }

        return $EMail;
    }


    /**
     * Fetch the headers of an email from backend
     *
     * @param string $archive
     * @param string $sessionMailId
     * @return array $headers
     */
    function fetchHeaders ($cidstring,$sessionMailId)
    {
        $bennoId = QueryResult::getBennoId($sessionMailId);

        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);
        $query_string = 'archive='.$archive.
                        '&id='.$bennoId.
                        '&format=headers';
        if ($scid)  $query_string .= '&scid='.$scid;

        $header = $this->backend_request('/mail',$query_string,'headers');

        return $header;
    }


    /**
     * Fetch one email in raw format from backend
     *
     * @param string $archive
     * @param string $sessionMailId
     * @return string $maildata
     */
    function fetchRaw ($cidstring,$sessionMailId,$recodeToUTF8=false)
    {
        $bennoId = QueryResult::getBennoId($sessionMailId);

        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);
        $query_string = 'archive='.$archive.
                        '&id='.$bennoId.
                        '&format=raw';
        if ($scid)  $query_string .= '&scid='.$scid;

        if (!$recodeToUTF8) {
            $query_string .= '&skipUTF8Recode=true';
        }

        $maildata = $this->backend_request('/mail',$query_string,'raw');

        return $maildata;
    }


    /**
     * Fetch attachment from backend
     *
     * @param string $archive
     * @param string $mailId
     * @return object $EMail
     */
    function fetchAttachment ($cidstring,$sessionMailId,$filename, $attachmentId = '')
    {
        $bennoId = QueryResult::getBennoId($sessionMailId);

        $request_url  = $this->base_url.'/attachment/';
        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);
        $query_string = 'archive='.$archive.
                        '&id='.$bennoId.
                        '&attachment='.$filename;
        if ($scid)  $query_string .= '&scid='.$scid;
                        
	    if($attachmentId != "")         
	      $query_string .= '&attachmentId='.$attachmentId;

        $file = $this->backend_request('/attachment',$query_string,'raw');
        if (! $file) {
            throw new Exception('NO_ATTACHMENT');
        }

        return $file;
    }


    /**
     * Fetch license data from backend
     *
     * @return object $License
     */
    function fetchLicense()
    {
        global $App;

        $data = $this->backend_request('/license');

        if (! $data) {
            $App->log('[Backend/fetchLicense] No valid License found');
        }

        $License = new License($data['licenseStatus'],$data['licenseExpires'],$data['licenseType'],$data['licenseFile'],$data['myAddress'],$data['myHostname'],$data['archiveCount'],$data['coreVersion'],$data['restVersion'],$data['licenseInstanceId']);

        return $License;
    }



    /**
     * Search backend for mails
     *
     * @param object $Query
     * @param start $startnum (default: 1)
     * @param int $limit (default: 20)
     * @return array $EMail objects
     */
    function query ($cidstring,$Query,$startNum=0,$limit=20)
    {
        global $App;
        global $Audit;

        if (empty($this->filter)) {
            $this->filter = 'NOT*';
        }

        $query = $Query->getQuery();
        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);  // scid: optional extension of name
        $query_string = 'archive='.$archive .
                 '&filterQuery='.$this->filter .
                 '&limit='.$limit .
                 '&start='.$startNum .
                 '&sort='.$Query->getSortField() .
                 '&sortAsc='.$Query->getSortAsc() .
                 '&query='.$query;
        if ($scid)  $query_string .= '&scid='.$scid;

        $App->log("Query string: $query",3,'Backend');
        $App->log("Filter string: $this->filter",4,'Backend');

        $resultList = $this->backend_request('/search',$query_string);

        if (! preg_match('/^LICENSE_VALID/',$resultList['licenseValid'])) {
            if ($App->getConfig('USERPERMISSONS','ENABLED') != 'DISABLED') {
                $App->setMsg('LICENSE_ERROR');
                throw new Exception('LICENSE_ERROR');
            }
        }
        $Audit->log("QUERY $archive:".$Query->getQuery().' && '.$this->filter);

        $QueryResult = new QueryResult($resultList,$Query,$limit);
        return $QueryResult;
    }


    /**
     * Backend request with fallback
     *
     * @param string $endpoint
     * @param string $query=false
     * @param string $format=''
     * @return array
     */
    function backend_request($endpoint,$query_string=false,$format='')
    {
        global $App;
        $maildata = '';

        $request_url  = $this->base_url.$endpoint.'/';

        try {
            $maildata = $this->_curl_query($request_url,$query_string,$format);
        }
        catch (exception $e) {  // FS#1105
            $error_msg = $e->getMessage();
            if ($error_msg == 'ERR_REST_404') {
                $base_url_fallback = $App->getConfig('BASE_URL_FALLBACK');
                if ($base_url_fallback) {
                    $fallback_url = rtrim($base_url_fallback,'/').$endpoint.'/';
                    $App->log('Try to fetch from fallback backend: '.$fallback_url,2,'Backend');
                    $maildata = $this->_curl_query($fallback_url,$query_string,$format);
                }
                else {
                    throw new Exception('ERR_REST_404');
                }
            }
            else {
                throw $e;
            }
        }

        return $maildata;
    }
 
    /**
     * Sends http query to benno2rest backend server.
     *
     * @param string $request_url
     * @param string $query=false
     * @param string $format=''
     * @return array
     */
    function _curl_query ($request_url,$query=false,$format='')
    {
        global $App;
        $cookie_file = $App->getConfig('COOKIE_FILE','/var/tmp/.benno.cookie')
                      .$App->getSession('USER')->id;

        if (!$this->_curl) {
            if (extension_loaded('curl')) {
                $this->_curl = curl_init();
            }
            else {
                $App->log('ERROR: curl module not installed!');
                throw new Exception('ERR_INTERN');
            }
        }

        $curl =& $this->_curl;


        $rest_user = $this->rest_user;
        $rest_pass = $this->rest_pass; 
        $http_header = array('Expect:');

        $url_parts = parse_url($request_url);
        if (array_key_exists('user',$url_parts) && array_key_exists('pass',$url_parts)) {
            if (preg_match('/^Bearer$/',$url_parts['user'])) {        // set Bearer token
                //array_push($http_header,'Authorization: Bearer '.$this->gen_jwt($url_parts['pass']));  // gen_jwt() not implemented
                $App->log('WARNING: JWT token handling not implemented yet',0,'Backend');
            }
            else {
                $rest_user = $url_parts['user'];
                $rest_user = $url_parts['user'];
                array_push($http_header,'Authorization: Basic '.base64_encode($rest_user.':'.$rest_pass));
            }
        }
        curl_setopt($curl, CURLOPT_USERPWD, $rest_user.':'.$rest_pass); 

        curl_setopt($curl, CURLOPT_URL, $request_url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($curl, CURLOPT_HEADER, 1);
        curl_setopt($curl, CURLOPT_POST, 1);
        curl_setopt($curl, CURLOPT_POSTFIELDS, $query);
        // suppress using of 100-Continue requests
        curl_setopt($curl, CURLOPT_HTTPHEADER,$http_header);
        if ($cookie_file) {
            curl_setopt($curl, CURLOPT_COOKIEJAR, $cookie_file);
            curl_setopt($curl, CURLOPT_COOKIEFILE, $cookie_file);
        }
        if (preg_match('/^https:/',$request_url)) {
            if ($this->rest_ssl_verify == 'YES') {
                curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, true);
                curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, true);
            }
            else {
                curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
                curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, false);
            }
        }

        // set proxy for cURL requests (null or value from config / environment)
        $rest_proxy = strtolower($App->getConfig('REST_PROXY',''));
        if ($rest_proxy != 'environment') {
            curl_setopt($curl, CURLOPT_PROXY, $rest_proxy);
        }

        $App->log('Request URL: '.$request_url."?".$query,4,'Backend');
        if (($response = curl_exec($curl)) === false) {
            $errnum = curl_errno($curl);
            $App->log('[Backend/query] Request error: '.curl_error($curl).'['.$errnum.']');
            if ($errnum == 7) {
                throw new Exception('ERR_REST_CON');
            }
            else {
                throw new Exception('ERR_REQUEST');
            }
        }
        $httpCode = curl_getinfo($curl, CURLINFO_HTTP_CODE); 
        list($response_headers, $response_body) = explode("\r\n\r\n", $response, 2);

        switch ($httpCode) {
            case 200:
                break;
            case 401:
                $App->log('[Backend/query] HTTP STATUS '.$httpCode.', REST backed password incorrect.');
                throw new Exception('ERR_REST_'.$httpCode);
            case 404:
                $jbody = json_decode($response_body,true);
                if ($jbody['error'] == "container not initialized") {
                    $App->log("[Backend/query] HTTP STATUS $httpCode, Container '".$jbody['container']."' not found");
                    throw new Exception('ERR_INITIALIZE');
                }
                elseif ($jbody['error'] == "container not found") {
                    $App->log("[Backend/query] HTTP STATUS $httpCode, Container not found");
                    throw new Exception('ERR_ARCHIVE');
                }
                else {
                    throw new Exception('ERR_REST_404');
                }
            default:
                $App->log('[Backend/query] HTTP ERROR '.$httpCode.', query: "'.$query.'"');
                throw new Exception('ERR_REST_'.$httpCode);
        }

        if (!$response_body) {
            // empty body (logout,...) 
            $App->log('Empty response from backend',4,'Backend');
            return '';
        }

        if ($format == 'raw') {
            $result = $response_body;
        }
        else {
            if (!mb_check_encoding($response_body, 'UTF-8') ) {
                $response_body = mb_convert_encoding($response_body, 'UTF-8');
            }
            $result = json_decode($response_body, true);

            if (array_key_exists('error',$result)) {
                $App->log('[Backend/query] '.$result['error']);
                throw new Exception($result['error']);
            }
        }

        return $result;
    }

    function destroy_bennorest_session()
    {
        $this->backend_request('/logout');
    }

}
?>

==> benno-web-2.10.8.tgz: srv/benno/bennoweb/htdocs/model/Backend.php <==
<?php
/*
 * Benno MailArchiv
 *
 * Copyright  2008-2012 LWsystems GmbH & Co. KG
 *
 * http://www.lw-systems.de/
 * http://www.benno-mailarchiv.de/
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License version 3
 * as published by the Free Software Foundation.
 *
 * Binary versions of this file provided by LWsystems to you as
 * well as other copyrighted, protected or trademarked materials like
 * logos, graphics, fonts, specific documentations and configurations,
 * cryptographic keys etc. are subject to a license agreement between
 * you and LWsystems.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You can find a copy of the GNU Affero General Public License at
 * this URI:
 * http://www.gnu.org/licenses/agpl-3.0.html
 * If not, write to the Free Software Foundation, Inc., 59 Temple Place,
 * Suite 330, Boston, MA  02111-1307  USA
 */

require_once 'model/EMail.php';
require_once 'model/Attachment.php';
require_once 'model/QueryResult.php';
require_once 'model/License.php';
require_once 'model/ContainerBackend.php';

class Backend
{

    var $base_url;
    var $filter;        // filter assign to each query
    var $rest_user;
    var $rest_pass;
    var $rest_ssl_verify;
    var $_curl;

    function __construct($base_url,$rest_user,$rest_pass,$filter)
    {
        global $App;
        $this->rest_user = $rest_user;
        $this->rest_pass = $rest_pass;
        $this->rest_ssl_verify = $App->getConfig('REST_SSL_VERIFY','NO');

        $this->base_url = $base_url;
        $this->filter = $filter;
    }



    function __serialize()
    {
        $odata['base_url']        = $this->base_url;
        $odata['filter']          = $this->filter;
        $odata['rest_pass']       = $this->rest_pass;
        $odata['rest_ssl_verify'] = $this->rest_ssl_verify;
        $odata['rest_user']       = $this->rest_user;

        return $odata;
    }


    function __unserialize($odata)
    {
        $this->base_url         = $odata['base_url'];
        $this->filter           = $odata['filter'];
        $this->rest_pass        = $odata['rest_pass'];
        $this->rest_ssl_verify  = $odata['rest_ssl_verify'];
        $this->rest_user        = $odata['rest_user'];

        $this->_curl = curl_init();
    }


    /**
     * Fetch one email from backend
     *
     * @param string $archive
     * @param string $sessionMailId
     * @return object $EMail
     */
    function fetch ($cidstring,$sessionMailId)
    {
        global $App;
        $bennoId = QueryResult::getBennoId($sessionMailId);
        $App->log("Fetch mail from archive: $cidstring/$bennoId",3,'Backend');

        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);
        $query_string = 'archive='.$archive.
                        '&id='.$bennoId;
        if ($scid)  $query_string .= '&scid='.$scid;

        $maildata = $this->backend_request($archive,'/mail',$query_string);

        $body   = $maildata['body'];
        $header = $maildata['header'];

        $EMail = new EMail($sessionMailId,$archive,$bennoId,$header['From'],$header['To'],$header['Cc'],$header['Bcc'],$header['Subject'],$header['Date'],$header['Size'],$body);
        $EMail->verified = $header['X-Benno-HashCheckPassed'];
        $EMail->dateTS   = $header['X-Benno-DateTS'];

        foreach ($maildata['attachments'] as $atm) {
            //$Attachment = new Attachment($atm['name'],@$atm['content-type'], $atm['size']);
        	$Attachment = new Attachment($atm['name'],@$atm['content-type'], $atm['size'], true, $atm['attachmentId']);
            $EMail->addAttachment($Attachment);
        }

        return $EMail;
    }


    /**
     * Fetch the headers of an email from backend
     *
     * @param string $archive
     * @param string $sessionMailId
     * @return array $headers
     */
    function fetchHeaders ($cidstring,$sessionMailId)
    {
        $bennoId = QueryResult::getBennoId($sessionMailId);

        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);
        $query_string = 'archive='.$archive.
                        '&id='.$bennoId.
                        '&format=headers';
        if ($scid)  $query_string .= '&scid='.$scid;

        $header = $this->backend_request($archive,'/mail',$query_string,'headers');

        return $header;
    }


    /**
     * Fetch one email in raw format from backend
     *
     * @param string $archive
     * @param string $sessionMailId
     * @return string $maildata
     */
    function fetchRaw ($cidstring,$sessionMailId,$recodeToUTF8=false)
    {
        $bennoId = QueryResult::getBennoId($sessionMailId);

        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);
        $query_string = 'archive='.$archive.
                        '&id='.$bennoId.
                        '&format=raw';
        if ($scid)  $query_string .= '&scid='.$scid;

        if (!$recodeToUTF8) {
            $query_string .= '&skipUTF8Recode=true';
        }

        $maildata = $this->backend_request($archive,'/mail',$query_string,'raw');

        return $maildata;
    }


    /**
     * Fetch attachment from backend
     *
     * @param string $archive
     * @param string $mailId
     * @return object $EMail
     */
    function fetchAttachment ($cidstring,$sessionMailId,$filename, $attachmentId = '')
    {
        $bennoId = QueryResult::getBennoId($sessionMailId);

        $request_url  = $this->base_url.'/attachment/';
        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);
        $query_string = 'archive='.$archive.
                        '&id='.$bennoId.
                        '&attachment='.$filename;
        if ($scid)  $query_string .= '&scid='.$scid;
                        
	    if($attachmentId != "")         
	      $query_string .= '&attachmentId='.$attachmentId;

        $file = $this->backend_request($archive,'/attachment',$query_string,'raw');
        if (! $file) {
            throw new Exception('NO_ATTACHMENT');
        }

        return $file;
    }


    /**
     * Fetch license data from backend
     *
     * @param string $archive
     * @return object $License
     */
    function fetchLicense($archive)
    {
        global $App;

        $data = $this->backend_request($archive,'/license');

        if (! $data) {
            $App->log('[Backend/fetchLicense] No valid License found');
        }

        $License = new License($data['licenseStatus'],$data['licenseExpires'],$data['licenseType'],$data['licenseFile'],$data['myAddress'],$data['myHostname'],$data['archiveCount'],$data['coreVersion'],$data['restVersion'],$data['licenseInstanceId']);

        return $License;
    }



    /**
     * Search backend for mails
     *
     * @param object $Query
     * @param start $startnum (default: 1)
     * @param int $limit (default: 20)
     * @return array $EMail objects
     */
    function query ($cidstring,$Query,$startNum=0,$limit=20)
    {
        global $App;
        global $Audit;

        if (empty($this->filter)) {
            $this->filter = 'NOT*';
        }

        $query = $Query->getQuery();
        @list($archive,$scid) = preg_split('/\/+(?=[^\/]+$)/',$cidstring);  // scid: optional extension of name
        $query_string = 'archive='.$archive .
                 '&filterQuery='.$this->filter .
                 '&limit='.$limit .
                 '&start='.$startNum .
                 '&sort='.$Query->getSortField() .
                 '&sortAsc='.$Query->getSortAsc() .
                 '&query='.$query;
        if ($scid)  $query_string .= '&scid='.$scid;

        $App->log("Query string: $query",3,'Backend');
        $App->log("Filter string: $this->filter",4,'Backend');

        $resultList = $this->backend_request($archive,'/search',$query_string);

        if (! preg_match('/^LICENSE_VALID/',$resultList['licenseValid'])) {
            if ($App->getConfig('USERPERMISSONS','ENABLED') != 'DISABLED') {
                $App->setMsg('LICENSE_ERROR');
                throw new Exception('LICENSE_ERROR');
            }
        }
        $Audit->log("QUERY $archive:".$Query->getQuery().' && '.$this->filter);

        $QueryResult = new QueryResult($resultList,$Query,$limit);
        return $QueryResult;
    }


    /**
     * Backend request with fallback
     *
     * @param string $endpoint
     * @param string $query=false
     * @param string $format=''
     * @return array
     */
    function backend_request($archive,$endpoint,$query_string=false,$format='')
    {
        global $App;
        $maildata = '';


        $cbackends = [];
        try {
            $cbackends = ContainerBackend::getBackends($archive);
        }
        catch (exception $e) {  // table will used on complex cloud setups
            $App->log('Backend table containerbackend not initialized',4,'Backend');
        }

        $request_url  = $this->base_url;
        $auth_type = 'Basic';
        $auth_data = base64_encode($this->rest_user.':'.$this->rest_pass);
        array_push($cbackends,new ContainerBackend($archive,$request_url,$auth_type,$auth_data));    // set default backend

        $error_sym = '';
        foreach ($cbackends as $cbackend) {
            try {

                $http_header = array();
                $auth_header = 'Authorization: '.$cbackend->auth_type().' '.$cbackend->auth_data();
                $request_url = $cbackend->url().$endpoint.'/';

                array_push($http_header,$auth_header);
                $maildata = $this->_curl_query($request_url,$http_header,$query_string,$format);

                return $maildata;
            }
            catch (exception $e) {  // FS#1105
                $error_sym = $e->getMessage();
                if ($error_sym == 'ERR_REST_404') {
                    $App->log('Email not found at backend: '.$request_url,2,'Backend');
                    continue;
                }
                else {
                    throw new Exception($error_sym);
                }
            }
        }
        if ($error_sym) {
            throw new Exception($error_sym);
        }

    }
 
    /**
     * Sends http query to benno2rest backend server.
     *
     * @param string $request_url
     * @param string $query=false
     * @param string $format=''
     * @return array
     */
    function _curl_query ($request_url,$http_header,$query=false,$format='')
    {
        global $App;
        $cookie_file = $App->getConfig('COOKIE_FILE','/var/tmp/.benno.cookie').@$App->getSession('USER')->id;

        if (!$this->_curl) {
            if (extension_loaded('curl')) {
                $this->_curl = curl_init();
            }
            else {
                $App->log('ERROR: curl module not installed!');
                throw new Exception('ERR_INTERN');
            }
        }

        $curl =& $this->_curl;

        array_push($http_header,'Expect:');

        //curl_setopt($curl, CURLOPT_USERPWD, $rest_user.':'.$rest_pass); 

        curl_setopt($curl, CURLOPT_URL, $request_url);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($curl, CURLOPT_HEADER, 1);
        curl_setopt($curl, CURLOPT_POST, 1);
        curl_setopt($curl, CURLOPT_POSTFIELDS, $query);
        // suppress using of 100-Continue requests
        curl_setopt($curl, CURLOPT_HTTPHEADER,$http_header);
        if ($cookie_file) {
            curl_setopt($curl, CURLOPT_COOKIEJAR, $cookie_file);
            curl_setopt($curl, CURLOPT_COOKIEFILE, $cookie_file);
        }
        if (preg_match('/^https:/',$request_url)) {
            if ($this->rest_ssl_verify == 'YES') {
                curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, true);
                curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, true);
            }
            else {
                curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
                curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, false);
            }
        }

        // set proxy for cURL requests (null or value from config / environment)
        $rest_proxy = strtolower($App->getConfig('REST_PROXY',''));
        if ($rest_proxy != 'environment') {
            curl_setopt($curl, CURLOPT_PROXY, $rest_proxy);
        }

        $App->log('Request URL: '.$request_url."?".$query,4,'Backend');
        if (($response = curl_exec($curl)) === false) {
            $errnum = curl_errno($curl);
            $App->log('[Backend/query] Request error: '.curl_error($curl).'['.$errnum.']');
            if ($errnum == 7) {
                throw new Exception('ERR_REST_CON');
            }
            else {
                throw new Exception('ERR_REQUEST');
            }
        }
        $httpCode = curl_getinfo($curl, CURLINFO_HTTP_CODE); 
        list($response_headers, $response_body) = explode("\r\n\r\n", $response, 2);

        switch ($httpCode) {
            case 200:
                break;
            case 401:
                $App->log('[Backend/query] HTTP STATUS '.$httpCode.', REST backed password incorrect.');
                throw new Exception('ERR_REST_'.$httpCode);
            case 404:
                $jbody = json_decode($response_body,true);
                if ($jbody['error'] == "container not initialized") {
                    $App->log("[Backend/query] HTTP STATUS $httpCode, Container '".$jbody['container']."' not found");
                    throw new Exception('ERR_INITIALIZE');
                }
                elseif ($jbody['error'] == "container not found") {
                    $App->log("[Backend/query] HTTP STATUS $httpCode, Container not found");
                    throw new Exception('ERR_ARCHIVE');
                }
                else {
                    throw new Exception('ERR_REST_404');
                }
            default:
                $App->log('[Backend/query] HTTP ERROR '.$httpCode.', query: "'.$query.'"');
                throw new Exception('ERR_REST_'.$httpCode);
        }

        if (!$response_body) {
            // empty body (logout,...) 
            $App->log('Empty response from backend',4,'Backend');
            return '';
        }

        if ($format == 'raw') {
            $result = $response_body;
        }
        else {
            if (!mb_check_encoding($response_body, 'UTF-8') ) {
                $response_body = mb_convert_encoding($response_body, 'UTF-8');
            }
            $result = json_decode($response_body, true);

            if (array_key_exists('error',$result)) {
                $App->log('[Backend/query] '.$result['error']);
                throw new Exception($result['error']);
            }
        }

        return $result;
    }

    function destroy_bennorest_session($archive)
    {
        $this->backend_request($archive,'/logout');
    }

}
?>
